using System;
using System.Diagnostics;
using System.Drawing;
using System.Globalization;
using System.Runtime.InteropServices;
using System.Windows.Forms;

using Microsoft.VisualStudio.Shell.Interop;

using ErrorHandler = Microsoft.VisualStudio.ErrorHandler;
using VSConstants = Microsoft.VisualStudio.VSConstants;

namespace Banquet
{
	[ComVisible(true), Guid(GuidList.guidSccProviderService)]
	internal sealed class SccProviderService : IVsSccProvider, IVsSccManager2, IVsSccGlyphs, IVsSccManagerTooltip, IVsQueryEditQuerySave2, IVsSolutionEvents2, IVsTrackProjectDocumentsEvents2, IDisposable
	{
		#region Fields

		private readonly IServiceProvider serviceProvider;

		private bool active;

		private uint vsSolutionEventsCookie;
		private uint tpdTrackProjectDocumentsCookie;

		private ImageList glyphs;

		#endregion

		#region Events

		public event EventHandler<EventArgs> ActiveStateChanged;

		#endregion

		#region Constructors

		public SccProviderService(IServiceProvider serviceProvider)
		{
			Debug.Assert(serviceProvider != null);
			this.serviceProvider = serviceProvider;

			IVsSolution solution = (IVsSolution)serviceProvider.GetService(typeof(SVsSolution));
			solution.AdviseSolutionEvents(this, out this.vsSolutionEventsCookie);
			Debug.Assert(this.vsSolutionEventsCookie != VSConstants.VSCOOKIE_NIL);

			IVsTrackProjectDocuments2 tpdService = (IVsTrackProjectDocuments2)serviceProvider.GetService(typeof(SVsTrackProjectDocuments));
			tpdService.AdviseTrackProjectDocumentsEvents(this, out this.tpdTrackProjectDocumentsCookie);
			Debug.Assert(this.tpdTrackProjectDocumentsCookie != VSConstants.VSCOOKIE_NIL);
		}

		#endregion

		#region Methods

		public void Dispose()
		{
			if (this.vsSolutionEventsCookie != VSConstants.VSCOOKIE_NIL)
			{
				IVsSolution solution = (IVsSolution)this.serviceProvider.GetService(typeof(SVsSolution));
				solution.UnadviseSolutionEvents(this.vsSolutionEventsCookie);
				this.vsSolutionEventsCookie = VSConstants.VSCOOKIE_NIL;
			}

			if (this.tpdTrackProjectDocumentsCookie != VSConstants.VSCOOKIE_NIL)
			{
				IVsTrackProjectDocuments2 tpdService = (IVsTrackProjectDocuments2)this.serviceProvider.GetService(typeof(SVsTrackProjectDocuments));
				tpdService.UnadviseTrackProjectDocumentsEvents(this.tpdTrackProjectDocumentsCookie);
				this.tpdTrackProjectDocumentsCookie = VSConstants.VSCOOKIE_NIL;
			}
		}

		#endregion

		#region Properties

		public bool Active
		{
			get
			{
				return this.active;
			}
		}

		#endregion

		#region IVsSccProvider Members

		public int AnyItemsUnderSourceControl(out int pfResult)
		{
			pfResult = 0;
			return VSConstants.S_OK;
		}

		public int SetActive()
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "Provider set active"));

			this.active = true;

			EventHandler<EventArgs> handler = this.ActiveStateChanged;
			if (handler != null)
			{
				handler(this, EventArgs.Empty);
			}

			return VSConstants.S_OK;
		}

		public int SetInactive()
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "Provider set inactive"));

			this.active = false;

			EventHandler<EventArgs> handler = this.ActiveStateChanged;
			if (handler != null)
			{
				handler(this, EventArgs.Empty);
			}

			return VSConstants.S_OK;
		}

		#endregion

		#region IVsSccManager2 Members

		public int BrowseForProject(out string pbstrDirectory, out int pfOK)
		{
			pbstrDirectory = null;
			pfOK = 0;
			return VSConstants.E_NOTIMPL;
		}

		public int CancelAfterBrowseForProject()
		{
			return VSConstants.E_NOTIMPL;
		}

		// TODO: GetSccGlyph
		public int GetSccGlyph(int cFiles, string[] rgpszFullPaths, VsStateIcon[] rgsiGlyphs, uint[] rgdwSccStatus)
		{
			rgsiGlyphs[0] = VsStateIcon.STATEICON_MAXINDEX;
			if (rgdwSccStatus != null)
			{
				rgdwSccStatus[0] = 99;
			}
			return VSConstants.S_OK;
		}

		// TODO: GetSccGlyphFromStatus
		public int GetSccGlyphFromStatus(uint dwSccStatus, VsStateIcon[] psiGlyph)
		{
			switch (dwSccStatus)
			{
			case 99:
				psiGlyph[0] = VsStateIcon.STATEICON_MAXINDEX;
				break;
			}
			return VSConstants.S_OK;
		}

		public int IsInstalled(out int pbInstalled)
		{
			pbInstalled = 1;
			return VSConstants.S_OK;
		}

		// TODO: RegisterSccProject
		public int RegisterSccProject(IVsSccProject2 pscp2Project, string pszSccProjectName, string pszSccAuxPath, string pszSccLocalPath, string pszProvider)
		{
			if (pscp2Project == null)
			{
				Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "Solution {0} is registering with source control - {1}, {2}, {3}, {4}", null, pszSccProjectName, pszSccAuxPath, pszSccLocalPath, pszProvider));
			}
			else
			{
				Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "Project {0} is registering with source control - {1}, {2}, {3}, {4}", null, pszSccProjectName, pszSccAuxPath, pszSccLocalPath, pszProvider));
			}

			return VSConstants.S_OK;
		}

		// TODO: UnregisterSccProject
		public int UnregisterSccProject(IVsSccProject2 pscp2Project)
		{
			if (pscp2Project == null)
			{
				Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "Solution {0} is unregistering with source control.", null));
			}
			else
			{
				Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "Project {0} is unregistering with source control.", null));
			}

			return VSConstants.S_OK;
		}

		#endregion

		#region IVsSccGlyphs Members

		public int GetCustomGlyphList(uint BaseIndex, out uint pdwImageListHandle)
		{
			if (this.glyphs == null)
			{
				IVsResourceManager resources = (IVsResourceManager)this.serviceProvider.GetService(typeof(SVsResourceManager));
				IUIHostLocale loc = (IUIHostLocale)this.serviceProvider.GetService(typeof(IUIHostLocale));

				uint culture;
				IntPtr hbitmap;
				Guid guidPackage = new Guid(GuidList.guidPackage);
				ErrorHandler.ThrowOnFailure(loc.GetUILocale(out culture));
				ErrorHandler.ThrowOnFailure(resources.LoadResourceBitmap(ref guidPackage, (int)culture, "@303", out hbitmap));

				this.glyphs = new ImageList();
				this.glyphs.ColorDepth = ColorDepth.Depth24Bit;
				this.glyphs.ImageSize = new Size(7, 16);
				this.glyphs.TransparentColor = Color.Magenta;
				this.glyphs.Images.AddStrip(Bitmap.FromHbitmap(hbitmap));
			}

			pdwImageListHandle = (uint)this.glyphs.Handle;
			return VSConstants.S_OK;
		}

		#endregion

		#region IVsSccManagerTooltip Members

		// TODO: GetGlyphTipText
		public int GetGlyphTipText(IVsHierarchy phierHierarchy, uint itemidNode, out string pbstrTooltipText)
		{
			pbstrTooltipText = string.Empty;
			return VSConstants.S_OK;
		}

		#endregion

		#region IVsSolutionEvents2 Members

		public int OnAfterCloseSolution(object pUnkReserved)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnAfterCloseSolution"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnAfterLoadProject(IVsHierarchy pStubHierarchy, IVsHierarchy pRealHierarchy)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnAfterLoadProject"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnAfterMergeSolution(object pUnkReserved)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnAfterMergeSolution"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnAfterOpenProject(IVsHierarchy pHierarchy, int fAdded)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnAfterOpenProject"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnAfterOpenSolution(object pUnkReserved, int fNewSolution)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnAfterOpenSolution"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnBeforeCloseProject(IVsHierarchy pHierarchy, int fRemoved)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnBeforeCloseProject"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnBeforeCloseSolution(object pUnkReserved)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnBeforeCloseSolution"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnBeforeUnloadProject(IVsHierarchy pRealHierarchy, IVsHierarchy pStubHierarchy)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnBeforeUnloadProject"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnQueryCloseProject(IVsHierarchy pHierarchy, int fRemoving, ref int pfCancel)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnQueryCloseProject"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnQueryCloseSolution(object pUnkReserved, ref int pfCancel)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnQueryCloseSolution"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnQueryUnloadProject(IVsHierarchy pRealHierarchy, ref int pfCancel)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnQueryUnloadProject"));

			return VSConstants.E_NOTIMPL;
		}

		#endregion

		#region IVsQueryEditQuerySave2 Members

		public int BeginQuerySaveBatch()
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "BeginQuerySaveBatch"));

			return VSConstants.S_OK;
		}

		public int DeclareReloadableFile(string pszMkDocument, uint rgf, VSQEQS_FILE_ATTRIBUTE_DATA[] pFileInfo)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "DeclareReloadableFile"));

			return VSConstants.S_OK;
		}

		public int DeclareUnreloadableFile(string pszMkDocument, uint rgf, VSQEQS_FILE_ATTRIBUTE_DATA[] pFileInfo)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "DeclareUnreloadableFile"));

			return VSConstants.S_OK;
		}

		public int EndQuerySaveBatch()
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "EndQuerySaveBatch"));

			return VSConstants.S_OK;
		}

		public int IsReloadable(string pszMkDocument, out int pbResult)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "IsReloadable"));

			pbResult = 1;
			return VSConstants.S_OK;
		}

		public int OnAfterSaveUnreloadableFile(string pszMkDocument, uint rgf, VSQEQS_FILE_ATTRIBUTE_DATA[] pFileInfo)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnAfterSaveUnreloadableFile"));

			return VSConstants.S_OK;
		}

		public int QueryEditFiles(uint rgfQueryEdit, int cFiles, string[] rgpszMkDocuments, uint[] rgrgf, VSQEQS_FILE_ATTRIBUTE_DATA[] rgFileInfo, out uint pfEditVerdict, out uint prgfMoreInfo)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "QueryEditFiles"));

			pfEditVerdict = (uint)tagVSQueryEditResult.QER_EditOK;
			prgfMoreInfo = 0;
			return VSConstants.S_OK;
		}

		public int QuerySaveFile(string pszMkDocument, uint rgf, VSQEQS_FILE_ATTRIBUTE_DATA[] pFileInfo, out uint pdwQSResult)
		{
			string[] rgszDocuments = new string[1];
			uint[] rgrgf = new uint[1];
			rgszDocuments[0] = pszMkDocument;
			rgrgf[0] = rgf;
			return QuerySaveFiles(rgf, 1, rgszDocuments, rgrgf, pFileInfo, out pdwQSResult);
		}

		public int QuerySaveFiles(uint rgfQuerySave, int cFiles, string[] rgpszMkDocuments, uint[] rgrgf, VSQEQS_FILE_ATTRIBUTE_DATA[] rgFileInfo, out uint pdwQSResult)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "QuerySaveFiles"));

			pdwQSResult = (uint)tagVSQuerySaveResult.QSR_SaveOK;
			return VSConstants.S_OK;
		}

		#endregion

		#region IVsTrackProjectDocumentsEvents2 Members

		public int OnAfterAddDirectoriesEx(int cProjects, int cDirectories, IVsProject[] rgpProjects, int[] rgFirstIndices, string[] rgpszMkDocuments, VSADDDIRECTORYFLAGS[] rgFlags)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnAfterAddDirectoriesEx"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnAfterAddFilesEx(int cProjects, int cFiles, IVsProject[] rgpProjects, int[] rgFirstIndices, string[] rgpszMkDocuments, VSADDFILEFLAGS[] rgFlags)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnAfterAddFilesEx"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnAfterRemoveDirectories(int cProjects, int cDirectories, IVsProject[] rgpProjects, int[] rgFirstIndices, string[] rgpszMkDocuments, VSREMOVEDIRECTORYFLAGS[] rgFlags)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnAfterRemoveDirectories"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnAfterRemoveFiles(int cProjects, int cFiles, IVsProject[] rgpProjects, int[] rgFirstIndices, string[] rgpszMkDocuments, VSREMOVEFILEFLAGS[] rgFlags)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnAfterRemoveFiles"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnAfterRenameDirectories(int cProjects, int cDirs, IVsProject[] rgpProjects, int[] rgFirstIndices, string[] rgszMkOldNames, string[] rgszMkNewNames, VSRENAMEDIRECTORYFLAGS[] rgFlags)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnAfterRenameDirectories"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnAfterRenameFiles(int cProjects, int cFiles, IVsProject[] rgpProjects, int[] rgFirstIndices, string[] rgszMkOldNames, string[] rgszMkNewNames, VSRENAMEFILEFLAGS[] rgFlags)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnAfterRenameFiles"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnAfterSccStatusChanged(int cProjects, int cFiles, IVsProject[] rgpProjects, int[] rgFirstIndices, string[] rgpszMkDocuments, uint[] rgdwSccStatus)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnAfterSccStatusChanged"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnQueryAddDirectories(IVsProject pProject, int cDirectories, string[] rgpszMkDocuments, VSQUERYADDDIRECTORYFLAGS[] rgFlags, VSQUERYADDDIRECTORYRESULTS[] pSummaryResult, VSQUERYADDDIRECTORYRESULTS[] rgResults)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnQueryAddDirectories"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnQueryAddFiles(IVsProject pProject, int cFiles, string[] rgpszMkDocuments, VSQUERYADDFILEFLAGS[] rgFlags, VSQUERYADDFILERESULTS[] pSummaryResult, VSQUERYADDFILERESULTS[] rgResults)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnQueryAddFiles"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnQueryRemoveDirectories(IVsProject pProject, int cDirectories, string[] rgpszMkDocuments, VSQUERYREMOVEDIRECTORYFLAGS[] rgFlags, VSQUERYREMOVEDIRECTORYRESULTS[] pSummaryResult, VSQUERYREMOVEDIRECTORYRESULTS[] rgResults)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnQueryRemoveDirectories"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnQueryRemoveFiles(IVsProject pProject, int cFiles, string[] rgpszMkDocuments, VSQUERYREMOVEFILEFLAGS[] rgFlags, VSQUERYREMOVEFILERESULTS[] pSummaryResult, VSQUERYREMOVEFILERESULTS[] rgResults)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnQueryRemoveFiles"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnQueryRenameDirectories(IVsProject pProject, int cDirs, string[] rgszMkOldNames, string[] rgszMkNewNames, VSQUERYRENAMEDIRECTORYFLAGS[] rgFlags, VSQUERYRENAMEDIRECTORYRESULTS[] pSummaryResult, VSQUERYRENAMEDIRECTORYRESULTS[] rgResults)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnQueryRenameDirectories"));

			return VSConstants.E_NOTIMPL;
		}

		public int OnQueryRenameFiles(IVsProject pProject, int cFiles, string[] rgszMkOldNames, string[] rgszMkNewNames, VSQUERYRENAMEFILEFLAGS[] rgFlags, VSQUERYRENAMEFILERESULTS[] pSummaryResult, VSQUERYRENAMEFILERESULTS[] rgResults)
		{
			Trace.WriteLine(String.Format(CultureInfo.CurrentUICulture, "OnQueryRenameFiles"));

			return VSConstants.E_NOTIMPL;
		}

		#endregion
	}
}