using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;

using Python;

namespace Canonical.Bazaar
{
	public sealed class Branch : IDisposable // TODO: catch PythonExceptions and throw subclasses of BazaarException
	{
		#region Fields

		private readonly PythonObject impl;

		#endregion

		#region Constructors

		internal Branch(PythonObject impl)
		{
			Debug.Assert(impl != null);
			Debug.Assert(!impl.IsInvalid);
			this.impl = impl;
		}

		#endregion

		~Branch()
		{
			Dispose(false);
		}

		public void Close()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool disposing)
		{
			if (disposing)
			{
				this.impl.Close();
			}
		}

		#region Properties

		public string DisplayName
		{
			get
			{
				using (PythonString nick = this.impl.GetAttribute<PythonString>("nick"))
				{
					return nick.ToString();
				}
			}
			set
			{
				using (PythonString nick = PythonString.CreateUnicode(value))
				{
					this.impl.SetAttribute("nick", nick);
				}
			}
		}

		public RevisionIdVersionSpec LatestRevisionId
		{
			get
			{
				throw new NotImplementedException(); // TODO
			}
		}

		public Uri Location
		{
			get
			{
				throw new NotImplementedException(); // TODO
			}
		}

		#endregion

		#region Static Methods

		public static Branch Inititialize(Uri location)
		{
			if (location == null)
			{
				throw new ArgumentNullException("location", Messages.ArgumentNull_Uri);
			}

			using (PythonString the_location = PythonString.CreateUnicode(location.AbsoluteUri))
			using (PythonModule bzrlib_transport = PythonModule.Import("bzrlib.transport"))
			using (PythonObject to_transport = bzrlib_transport.InvokeMethod("get_transport", the_location))
			using (to_transport.InvokeMethod("ensure_base"))
			using (PythonObject @base = to_transport.GetAttribute("base"))
			using (PythonModule bzrlib_bzrdir = PythonModule.Import("bzrlib.bzrdir"))
			using (PythonObject format_registry = bzrlib_bzrdir.GetAttribute("format_registry"))
			using (PythonString @default = PythonString.CreateString("default"))
			using (PythonObject format = format_registry.InvokeMethod("make_bzrdir", @default))
			using (PythonClass bzrlib_bzrdir_BzrDir = bzrlib_bzrdir.GetAttribute<PythonClass>("BzrDir"))
			{
				return new Branch(bzrlib_bzrdir_BzrDir.InvokeMethod("create_branch_convenience", @base, format));
			}
		}

		public static Branch Open(Uri location)
		{
			if (location == null)
			{
				throw new ArgumentNullException("location", Messages.ArgumentNull_Uri);
			}

			using (PythonString the_location = PythonString.CreateUnicode(location.AbsoluteUri))
			using (PythonModule bzrlib_branch = PythonModule.Import("bzrlib.branch"))
			using (PythonClass bzrlib_branch_Branch = bzrlib_branch.GetAttribute<PythonClass>("Branch"))
			{
				return new Branch(bzrlib_branch_Branch.InvokeMethod("open", the_location));
			}
		}

		public static Branch OpenContaining(Uri location)
		{
			if (location == null)
			{
				throw new ArgumentNullException("location", Messages.ArgumentNull_Uri);
			}

			using (PythonString the_location = PythonString.CreateUnicode(location.AbsoluteUri))
			using (PythonModule bzrlib_branch = PythonModule.Import("bzrlib.branch"))
			using (PythonClass bzrlib_branch_Branch = bzrlib_branch.GetAttribute<PythonClass>("Branch"))
			using (PythonTuple result = bzrlib_branch_Branch.InvokeMethod<PythonTuple>("open_containing", the_location))
			{
				return new Branch(result.GetItem(0));
			}
		}

		#endregion

		#region Methods

		public void Bind(Branch branch)
		{
			if (branch == null)
			{
				throw new ArgumentNullException("branch", Messages.ArgumentNull_Branch);
			}

			try
			{
				using (this.impl.InvokeMethod("bind", branch.impl))
				{
				}
			}
			catch (PythonException e)
			{
				if ("bzrlib.errors" == e.Type.Module && "DivergedBranches" == e.Type.Name)
				{
					throw new DivergedBranchesException();
				}
				else
				{
					throw;
				}
			}
		}

		public Branch Clone(Uri location)
		{
			throw new NotImplementedException(); // TODO
		}

		public void CreateTag(string tagName, VersionSpec revision)
		{
			if (tagName == null)
			{
				throw new ArgumentNullException("tagName", Messages.ArgumentNull_TagName);
			}
			if (revision == null)
			{
				throw new ArgumentNullException("revision", Messages.ArgumentNull_VersionSpec);
			}

			using (PythonString tag_name = PythonString.CreateUnicode(tagName))
			using (PythonObject tags = this.impl.GetAttribute("tags"))
			using (PythonObject has_tag = tags.InvokeMethod("has_tag", tag_name))
			{
				if (has_tag.IsTrue)
				{
					throw new TagAlreadyExistsException(tagName);
				}

				using (PythonString revision_id = PythonString.CreateString(null)) // TODO: get rev_id from versionspec
				using (tags.InvokeMethod("set_tag", tag_name, revision_id))
				{
				}
			}
		}

		public WorkingTree CreateWorkingTree(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}

			path = FileSpec.Canonicalize(Path.GetFullPath(path));

			if (this.Location.IsFile && string.Equals(path, this.Location.LocalPath, StringComparison.Ordinal))
			{
				using (PythonObject bzrdir = this.impl.GetAttribute("bzrdir"))
				{
					try
					{
						return new WorkingTree(bzrdir.InvokeMethod("open_workingtree"));
					}
					catch (PythonException e)
					{
						if ("bzrlib.errors" == e.Type.Module && "NoWorkingTree" == e.Type.Name)
						{
							return new WorkingTree(bzrdir.InvokeMethod("create_workingtree"));
						}
						else
						{
							throw;
						}
					}
				}
			}
			else
			{
				Directory.CreateDirectory(path);

				using (PythonString to_location = PythonString.CreateUnicode(path))
				using (PythonNone revision_id = PythonNone.Create())
				using (PythonBool lightweight = PythonBool.Create(false))
				{
					return new WorkingTree(this.impl.InvokeMethod("create_checkout", to_location, revision_id, lightweight));
				}
			}
		}

		public void DeleteTag(string tagName)
		{
			if (tagName == null)
			{
				throw new ArgumentNullException("tagName", Messages.ArgumentNull_TagName);
			}

			using (PythonString tag_name = PythonString.CreateUnicode(tagName))
			using (PythonObject tags = this.impl.GetAttribute("tags"))
			{
				try
				{
					using (tags.InvokeMethod("delete_tag", tag_name))
					{
					}
				}
				catch (PythonException e)
				{
					if ("bzrlib.errors" == e.Type.Module && "NoSuchTag" == e.Type.Name)
					{
						throw new NoSuchTagException(tagName);
					}
					else
					{
						throw;
					}
				}
			}
		}

		public void DeleteWorkingTree() // TODO: ensure branch is not a lightweight checkout
		{
			using (PythonObject bzrdir = this.impl.GetAttribute("bzrdir"))
			using (bzrdir.InvokeMethod("destroy_workingtree"))
			{
			}
		}

		public WorkingTree GetWorkingTree()
		{
			using (PythonObject bzrdir = this.impl.GetAttribute("bzrdir"))
			{
				try
				{
					return new WorkingTree(bzrdir.InvokeMethod("open_workingtree"));
				}
				catch (PythonException e)
				{
					if ("bzrlib.errors" == e.Type.Module && "NoWorkingTree" == e.Type.Name)
					{
						using (PythonString @base = e.Value.GetAttribute<PythonString>("base"))
						{
							throw new NoWorkingTreeException(@base.ToString());
						}
					}
					else
					{
						throw;
					}
				}
			}
		}

		public Branch GetMasterBranch()
		{
			throw new NotImplementedException(); // TODO
		}

		public void Pull(Uri location)
		{
			throw new NotImplementedException(); // TODO
		}

		public void Push(Uri location)
		{
			throw new NotImplementedException(); // TODO
		}

		public bool Unbind()
		{
			using (PythonObject result = this.impl.InvokeMethod("unbind"))
			{
				return result.IsTrue;
			}
		}

		public void Update()
		{
			using (this.impl.InvokeMethod<PythonInteger>("update")) // TODO: return result
			{
			}
		}

		#endregion
	}
}