using System;

using Python;

namespace Canonical.Bazaar
{
	public class Change
	{
		private readonly WorkingTree workingTree;
		private readonly string fileId;
		private readonly string oldPath;
		private readonly string path;
		private readonly bool contentChanged;
		private readonly bool oldVersioned;
		private readonly bool versioned;
		private readonly string oldParent;
		private readonly string parent;
		private readonly string oldName;
		private readonly string name;
		private readonly Kind? oldKind;
		private readonly Kind? kind;
		private readonly bool oldExecutable;
		private readonly bool executable;
		private readonly string oldAbsPath;
		private readonly string absPath;
		private readonly ChangeType type;

		internal Change(WorkingTree workingTree, PythonTuple change)
		{
			this.workingTree = workingTree;

			using (PythonString file_id = change.GetItem<PythonString>(0))
			{
				this.fileId = file_id.ToString();
			}

			using (PythonTuple path = change.GetItem<PythonTuple>(1))
			using (PythonString path_in_source = path.GetItem<PythonString>(0))
			using (PythonString path_in_target = path.GetItem<PythonString>(1))
			{
				this.oldPath = path_in_source.ToString();
				this.path = path_in_target.ToString();
			}

			using (PythonBool changed_content = change.GetItem<PythonBool>(2))
			{
				this.contentChanged = changed_content.IsTrue;
			}

			using (PythonTuple versioned = change.GetItem<PythonTuple>(3))
			using (PythonBool versioned_in_source = versioned.GetItem<PythonBool>(0))
			using (PythonBool versioned_in_target = versioned.GetItem<PythonBool>(1))
			{
				this.oldVersioned = versioned_in_source.IsTrue;
				this.versioned = versioned_in_target.IsTrue;
			}

			using (PythonTuple parent = change.GetItem<PythonTuple>(4))
			using (PythonString parent_in_source = parent.GetItem<PythonString>(0))
			using (PythonString parent_in_target = parent.GetItem<PythonString>(1))
			{
				this.oldParent = parent_in_source.ToString();
				this.parent = parent_in_target.ToString();
			}

			using (PythonTuple name = change.GetItem<PythonTuple>(5))
			using (PythonString name_in_source = name.GetItem<PythonString>(0))
			using (PythonString name_in_target = name.GetItem<PythonString>(1))
			{
				this.oldName = name_in_source.ToString();
				this.name = name_in_target.ToString();
			}

			using (PythonTuple kind = change.GetItem<PythonTuple>(6))
			using (PythonString kind_in_source = kind.GetItem<PythonString>(0))
			using (PythonString kind_in_target = kind.GetItem<PythonString>(1))
			{
				this.oldKind = GetKind(kind_in_source.ToString());
				this.kind = GetKind(kind_in_target.ToString());
			}

			using (PythonTuple executable = change.GetItem<PythonTuple>(7))
			using (PythonBool executable_in_source = executable.GetItem<PythonBool>(0))
			using (PythonBool executable_in_target = executable.GetItem<PythonBool>(1))
			{
				this.oldExecutable = executable_in_source.IsTrue;
				this.executable = executable_in_target.IsTrue;
			}

			if (this.oldPath != null)
				this.oldAbsPath = workingTree.GetAbsolutePath(this.oldPath);
			if (this.path != null)
				this.absPath = workingTree.GetAbsolutePath(this.path);

			if (this.oldName != null && this.name != null && this.oldParent != null && this.parent != null)
				if (this.oldName != this.name || this.oldParent != this.parent)
					this.type |= ChangeType.Rename;

			if (this.oldKind != this.kind)
				if (this.oldKind == null)
					this.type |= ChangeType.Create;
				else if (this.kind == null)
					this.type |= ChangeType.Delete;
				else
					this.type |= ChangeType.Kind;
			else
				if (this.contentChanged)
					this.type |= ChangeType.Edit;

			if (!this.oldVersioned)
				if (!this.versioned)
					this.type |= ChangeType.Unknown;
				else
					this.type |= ChangeType.Add;
			else
				if (!this.versioned)
					this.type |= ChangeType.Remove;

			//Console.WriteLine("{0,-17} {1}", this.type, this.path);
		}

		private static Kind? GetKind(string s)
		{
			switch (s)
			{
			case "file":
				return Kind.File;
			case "directory":
				return Kind.Directory;
			case "symlink":
				return Kind.SymbolicLink;
			case "tree-reference":
				return Kind.TreeReference;
			case null:
				return null;
			default:
				return Kind.Any;
			}
		}

		public static string[] ToAbsolutePaths(Change[] changes)
		{
			if (null == changes)
			{
				throw new ArgumentNullException("changes", Messages.ArgumentNull_Changes);
			}

			string[] result = new string[changes.Length];
			for (int i = 0; i < changes.Length; i++)
			{
				result[i] = changes[i].absPath;
			}
			return result;
		}

		public static string[] ToRelativePaths(Change[] changes)
		{
			if (null == changes)
			{
				throw new ArgumentNullException("changes", Messages.ArgumentNull_Changes);
			}

			string[] result = new string[changes.Length];
			for (int i = 0; i < changes.Length; i++)
			{
				result[i] = changes[i].path;
			}
			return result;
		}

		public ChangeType ChangeType
		{
			get
			{
				return this.type;
			}
		}

		public string FileName
		{
			get
			{
				return this.name;
			}
		}

		public bool IsAdd
		{
			get
			{
				return ((this.type & ChangeType.Add) != 0);
			}
		}

		public bool IsRemove
		{
			get
			{
				return ((this.type & ChangeType.Remove) != 0);
			}
		}

		public bool IsEdit
		{
			get
			{
				return ((this.type & ChangeType.Edit) != 0);
			}
		}

		public bool IsRename
		{
			get
			{
				return ((this.type & ChangeType.Rename) != 0);
			}
		}

		public string ItemId
		{
			get
			{
				return this.fileId;
			}
		}

		public Kind? ItemKind
		{
			get
			{
				return this.kind;
			}
		}

		public string FullPath
		{
			get
			{
				return this.absPath;
			}
		}

		public string Path
		{
			get
			{
				return this.path;
			}
		}
	}
}