using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;

using Python;

namespace Canonical.Bazaar
{
	public sealed class WorkingTree : IDisposable // TODO: catch PythonExceptions and throw subclasses of BazaarException
	{
		#region Fields

		private readonly PythonObject impl;
		private string basedir;

		#endregion

		#region Constructors

		internal WorkingTree(PythonObject impl)
		{
			Debug.Assert(impl != null);
			Debug.Assert(!impl.IsInvalid);
			this.impl = impl;
		}

		#endregion

		~WorkingTree()
		{
			Dispose(false);
		}

		public void Close()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool disposing)
		{
			if (disposing)
			{
				this.impl.Close();
				this.basedir = null;
			}
		}

		#region Properties

		public string BaseDirectory
		{
			get
			{
				if (this.basedir == null)
				{
					using (PythonString basedir = this.impl.GetAttribute<PythonString>("basedir"))
					{
						this.basedir = FileSpec.Canonicalize(Path.GetFullPath(basedir.ToString()));
					}
				}

				Debug.Assert(this.basedir != null);
				return this.basedir;
			}
		}

		#endregion

		#region Static Methods

		private static string[] MakeRelative(string[] items, string folder)
		{
			Debug.Assert(items != null);
			Debug.Assert(!string.IsNullOrEmpty(folder));
			Debug.Assert(string.Equals(folder, FileSpec.Canonicalize(Path.GetFullPath(folder)), StringComparison.Ordinal));

			string[] result = new string[items.Length];
			for (int i = 0; i < items.Length; i++)
			{
				result[i] = MakeRelative(FileSpec.Canonicalize(Path.GetFullPath(items[i])), folder);
			}
			return result;
		}

		private static string[] MakeAbsolute(string[] items, string folder)
		{
			Debug.Assert(items != null);
			Debug.Assert(!string.IsNullOrEmpty(folder));
			Debug.Assert(string.Equals(folder, FileSpec.Canonicalize(Path.GetFullPath(folder)), StringComparison.Ordinal));

			string[] result = new string[items.Length];
			for (int i = 0; items.Length > i; i++)
			{
				result[i] = MakeAbsolute(items[i], folder); // TODO: throw exception if string.IsNullOrEmpty(items[i])
			}
			return result;
		}

		private static string MakeRelative(string item, string folder)
		{
			Debug.Assert(!string.IsNullOrEmpty(item));
			Debug.Assert(string.Equals(item, FileSpec.Canonicalize(Path.GetFullPath(item)), StringComparison.Ordinal));
			Debug.Assert(!string.IsNullOrEmpty(folder));
			Debug.Assert(string.Equals(folder, FileSpec.Canonicalize(Path.GetFullPath(folder)), StringComparison.Ordinal));

			if (item.StartsWith(folder, StringComparison.Ordinal))
			{
				if (item.Length == folder.Length)
				{
					return string.Empty;
				}
				if (0 < folder.Length && Path.DirectorySeparatorChar == folder[folder.Length - 1])
				{
					return item.Substring(folder.Length).Replace(Path.DirectorySeparatorChar, '/');
				}
				if (Path.DirectorySeparatorChar == item[folder.Length])
				{
					return item.Substring(folder.Length + 1).Replace(Path.DirectorySeparatorChar, '/');
				}
			}

			throw new PathNotChildException(item, folder);
		}

		private static string MakeAbsolute(string item, string folder)
		{
			Debug.Assert(item != null);
			Debug.Assert(!string.IsNullOrEmpty(folder));
			Debug.Assert(string.Equals(folder, FileSpec.Canonicalize(Path.GetFullPath(folder)), StringComparison.Ordinal));

			return FileSpec.Canonicalize(Path.GetFullPath(Path.Combine(folder, item)));
		}

		public static WorkingTree Open(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}

			path = FileSpec.Canonicalize(Path.GetFullPath(path));

			using (PythonString the_path = PythonString.CreateUnicode(path))
			using (PythonModule bzrlib_workingtree = PythonModule.Import("bzrlib.workingtree"))
			using (PythonClass bzrlib_workingtree_WorkingTree = bzrlib_workingtree.GetAttribute<PythonClass>("WorkingTree"))
			{
				try
				{
					return new WorkingTree(bzrlib_workingtree_WorkingTree.InvokeMethod("open", the_path));
				}
				catch (PythonException e)
				{
					if ("bzrlib.errors" == e.Type.Module && "NotBranchError" == e.Type.Name)
					{
						throw new NotBranchException(path);
					}
					else
					{
						throw;
					}
				}
			}
		}

		public static WorkingTree OpenContaining(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}

			path = FileSpec.Canonicalize(Path.GetFullPath(path));

			/*
			using (PythonString the_path = PythonString.CreateUnicode(path))
			using (PythonModule bzrlib_workingtree = PythonModule.Import("bzrlib.workingtree"))
			using (PythonClass bzrlib_workingtree_WorkingTree = bzrlib_workingtree.GetAttribute<PythonClass>("WorkingTree"))
			using (PythonTuple result = bzrlib_workingtree_WorkingTree.InvokeMethod<PythonTuple>("open_containing", the_path))
			{
				return new WorkingTree(result.GetItem(0));
			}
			*/

			for (int index = path.Length - 1; index >= 0; index = path.LastIndexOf(Path.DirectorySeparatorChar, index - 1))
			{
				string test = path.Substring(0, index + 1);
				Debug.Assert(Path.IsPathRooted(test));
				if (Directory.Exists(Path.Combine(test, ".bzr")))
				{
					return Open(test);
				}
			}

			throw new NotBranchException(path);
		}

		private static void Recurse(string path, string folder, IList<string> items)
		{
			Debug.Assert(!string.IsNullOrEmpty(path));
			Debug.Assert(items != null);

			foreach (string item in Directory.GetFiles(path))
			{
				Debug.Assert(string.Equals(item, FileSpec.Canonicalize(Path.GetFullPath(item)), StringComparison.Ordinal));
				items.Add(MakeRelative(item, folder));
			}

			foreach (string item in Directory.GetDirectories(path))
			{
				Debug.Assert(string.Equals(item, FileSpec.Canonicalize(Path.GetFullPath(item)), StringComparison.Ordinal));
				items.Add(MakeRelative(item, folder));
				Recurse(item, folder, items);
			}
		}

		#endregion

		#region Methods

		public int Add(string[] paths)
		{
			return Add(paths, false);
		}

		public int Add(string path)
		{
			return Add(new string[] { path }, false);
		}

		public int Add(string[] paths, bool isRecursive)
		{
			if (paths == null)
			{
				throw new ArgumentNullException("paths", Messages.ArgumentNull_Paths);
			}
			if (paths.Length == 0)
			{
				return 0;
			}

			/*
			string[] the_paths = new string[paths.Length];
			for (int i = 0; i < paths.Length; i++)
			{
				the_paths[i] = Path.GetFullPath(paths[i]); // TODO: check paths[i] is in basedir
			}

			using (PythonList file_list = PythonList.Create(the_paths))
			using (PythonBool recurse = PythonBool.Create(isRecursive))
			using (PythonNone action = PythonNone.Create())
			using (PythonBool save = PythonBool.Create(true))
			using (PythonModule bzrlib_add = PythonModule.Import("bzrlib.add"))
			using (PythonTuple result = bzrlib_add.InvokeMethod<PythonTuple>("smart_add_tree", this.impl, file_list, recurse, action, save))
			using (PythonInteger added = result.GetItem<PythonInteger>(0))
			{
				return added.ToInt32();
			}
			*/

			this.impl.InvokeMethod("lock_tree_write");
			try
			{
				List<string> items = new List<string>();

				for (int i = 0; i < paths.Length; i++)
				{
					string path = FileSpec.Canonicalize(Path.GetFullPath(paths[i]));
					CollectUnversionedPaths(MakeRelative(path, BaseDirectory), items);
					if (isRecursive && Directory.Exists(path))
					{
						Recurse(path, BaseDirectory, items);
					}
				}

				if (items.Count == 0)
				{
					return 0;
				}

				using (PythonList files = PythonList.Create(items))
				using (this.impl.InvokeMethod("add", files))
				{
				}

				return items.Count;
			}
			finally
			{
				this.impl.InvokeMethod("unlock");
			}
		}

		public int Add(string path, bool isRecursive)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}
			if (path.Length == 0)
			{
				throw new ArgumentException(Messages.Argument_EmptyPath, "path");
			}

			return Add(new string[] { path }, isRecursive);
		}

		private void CollectUnversionedPaths(string path, IList<string> items)
		{
			Debug.Assert(!string.IsNullOrEmpty(path));
			Debug.Assert(path[path.Length - 1] != '/');
			Debug.Assert(items != null);

			if (!items.Contains(path))
			{
				using (PythonString the_path = PythonString.CreateUnicode(path))
				using (PythonObject id = this.impl.InvokeMethod("path2id", the_path))
				{
					if (id.IsFalse)
					{
						int i = path.LastIndexOf('/');
						if (i >= 0)
						{
							CollectUnversionedPaths(path.Substring(0, i), items);
						}
						items.Add(path);
					}
				}
			}
		}

		public void Commit(Change[] changes, string comment)
		{
			Commit(changes, null, comment);
		}

		public void Commit(Change[] changes, string author, string comment)
		{
			if (changes == null)
			{
				throw new ArgumentNullException("changes", Messages.ArgumentNull_Changes);
			}
			if (changes.Length == 0)
			{
				return;
			}

			string[] paths = Change.ToAbsolutePaths(changes); // TODO: check changes belong to workingtree

			using (PythonString message = PythonString.CreateUnicode(comment ?? string.Empty))
			using (PythonNone timestamp = PythonNone.Create())
			using (PythonNone timezone = PythonNone.Create())
			using (PythonObject committer = (author != null) ? (PythonObject)PythonString.CreateUnicode(author) : (PythonObject)PythonNone.Create())
			using (PythonList specific_files = PythonList.Create(paths))
			using (PythonNone rev_id = PythonNone.Create())
			using (PythonBool allow_pointless = PythonBool.Create(false))
			using (PythonBool strict = PythonBool.Create(false))
			using (PythonBool verbose = PythonBool.Create(false))
			using (PythonNone revprops = PythonNone.Create())
			using (PythonModule bzrlib_commit = PythonModule.Import("bzrlib.commit"))
			using (PythonClass bzrlib_commit_Commit = bzrlib_commit.GetAttribute<PythonClass>("Commit"))
			using (PythonObject commit = bzrlib_commit_Commit.Invoke())
			using (commit.InvokeMethod<PythonString>("commit", message, timestamp, timezone, committer, specific_files, rev_id, allow_pointless, strict, verbose, revprops, this.impl)) // TODO: return revision id
			{
			}
		}

		public void Delete(string[] paths)
		{
			if (paths == null)
			{
				throw new ArgumentNullException("paths", Messages.ArgumentNull_Paths);
			}
			if (paths.Length == 0)
			{
				return;
			}

			paths = MakeRelative(paths, BaseDirectory);

			using (PythonList files = PythonList.Create(paths))
			using (PythonBool verbose = PythonBool.Create(false))
			using (PythonNone to_file = PythonNone.Create())
			using (PythonBool keep_files = PythonBool.Create(false))
			using (PythonBool force = PythonBool.Create(true))
			using (this.impl.InvokeMethod("remove", files, verbose, to_file, keep_files, force))
			{
			}
		}

		public void Delete(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}

			Delete(new string[] { path });
		}

		public Change[] GetChanges()
		{
			List<Change> changes = new List<Change>();

			using (PythonObject from_tree = this.impl.InvokeMethod("basis_tree"))
			using (PythonBool include_unchanged = PythonBool.Create(false))
			using (PythonNone specific_files = PythonNone.Create())
			using (PythonNone pb = PythonNone.Create())
			using (PythonNone extra_trees = PythonNone.Create())
			using (PythonBool require_versioned = PythonBool.Create(true))
			using (PythonBool want_unversioned = PythonBool.Create(false))
			using (PythonIterable<PythonTuple> _iter_changes = this.impl.InvokeMethod<PythonIterable<PythonTuple>>("_iter_changes", from_tree, include_unchanged, specific_files, pb, extra_trees, require_versioned, want_unversioned))
			{
				foreach (PythonTuple change in _iter_changes)
				{
					changes.Add(new Change(this, change));
				}
			}

			return changes.ToArray();
		}

		public Conflict[] GetConflicts()
		{
			List<Conflict> conflicts = new List<Conflict>();

			using (PythonIterable<PythonObject> conflict_list = this.impl.InvokeMethod<PythonIterable<PythonObject>>("conflicts"))
			{
				foreach (PythonObject conflict in conflict_list)
				{
					using (PythonString path = conflict.GetAttribute<PythonString>("path"))
					using (PythonString typestring = conflict.GetAttribute<PythonString>("typestring"))
					{
						conflicts.Add(new Conflict(this, path.ToString(), typestring.ToString()));
					}
				}
			}

			return conflicts.ToArray();
		}

		public string GetAbsolutePath(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}

			/*
			using (PythonString filename = PythonString.CreateUnicode(path))
			using (PythonString abspath = this.impl.InvokeMethod<PythonString>("abspath", filename))
			{
				return Path.GetFullPath(abspath.ToString());
			}
			*/

			return MakeAbsolute(path, BaseDirectory);
		}

		public string GetRelativePath(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}
			// TODO: throw exception if (path.Length == 0)

			/*
			using (PythonString the_path = PythonString.CreateUnicode(Path.GetFullPath(path)))
			using (PythonString relpath = this.impl.InvokeMethod<PythonString>("relpath", the_path))
			{
				return relpath.ToString();
			}
			*/

			return MakeRelative(FileSpec.Canonicalize(Path.GetFullPath(path)), BaseDirectory);
		}

		public void Merge(Uri location)
		{
			throw new NotImplementedException(); // TODO
		}

		public void Rename(string oldPath, string newPath)
		{
			if (oldPath == null)
			{
				throw new ArgumentNullException("oldPath", Messages.ArgumentNull_Path);
			}
			if (newPath == null)
			{
				throw new ArgumentNullException("newPath", Messages.ArgumentNull_Path);
			}

			oldPath = MakeRelative(FileSpec.Canonicalize(Path.GetFullPath(oldPath)), BaseDirectory);
			newPath = MakeRelative(FileSpec.Canonicalize(Path.GetFullPath(newPath)), BaseDirectory);

			using (PythonString from_rel = PythonString.CreateUnicode(oldPath))
			using (PythonString to_rel = PythonString.CreateUnicode(newPath))
			using (this.impl.InvokeMethod("rename_one", from_rel, to_rel))
			{
			}
		}

		public void ResolveConflict(Conflict conflict)
		{
			if (conflict == null)
			{
				throw new ArgumentNullException("conflict", Messages.ArgumentNull_Conflict);
			}
			// TODO: check conflict belongs to workingtree

			using (PythonString path = PythonString.CreateUnicode(null)) // TODO: extract path from conflict
			using (PythonList paths = PythonList.Create(path))
			using (PythonBool ignore_misses = PythonBool.Create(false))
			using (PythonModule bzrlib_conflicts = PythonModule.Import("bzrlib.conflicts"))
			using (bzrlib_conflicts.InvokeMethod("resolve", this.impl, paths, ignore_misses))
			{
			}
		}

		public void Revert(string[] paths)
		{
			if (paths == null)
			{
				throw new ArgumentNullException("paths", Messages.ArgumentNull_Paths);
			}
			if (paths.Length == 0)
			{
				return;
			}

			paths = MakeRelative(paths, BaseDirectory);

			using (PythonList filenames = PythonList.Create(paths))
			using (PythonNone old_tree = PythonNone.Create())
			using (PythonBool backups = PythonBool.Create(false))
			using (this.impl.InvokeMethod<PythonList>("revert", filenames, old_tree, backups)) // TODO: return conflicts
			{
			}
		}

		public void Revert(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}

			Revert(new string[] { path });
		}

		public void Revert(string[] paths, VersionSpec revision)
		{
			if (paths == null)
			{
				throw new ArgumentNullException("paths", Messages.ArgumentNull_Paths);
			}
			if (revision == null)
			{
				throw new ArgumentNullException("revision", Messages.ArgumentNull_VersionSpec);
			}
			if (paths.Length == 0)
			{
				return;
			}

			paths = MakeRelative(paths, BaseDirectory);

			using (PythonList filenames = PythonList.Create(paths))
			using (PythonString rev_id = PythonString.CreateString(null)) // TODO: get rev_id from versionspec
			using (PythonObject branch = this.impl.GetAttribute("branch"))
			using (PythonObject repository = branch.GetAttribute("repository"))
			using (PythonObject old_tree = repository.InvokeMethod("revision_tree", rev_id))
			using (PythonBool backups = PythonBool.Create(false))
			using (this.impl.InvokeMethod<PythonList>("revert", filenames, old_tree, backups)) // TODO: return conflicts
			{
			}
		}

		public void Revert(string path, VersionSpec revision)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}

			Revert(new string[] { path }, revision);
		}

		public void Revert(Change change)
		{
			if (change == null)
			{
				throw new ArgumentNullException("change", Messages.ArgumentNull_Change);
			}

			Revert(new Change[] { change });
		}

		public void Revert(Change[] changes)
		{
			if (changes == null)
			{
				throw new ArgumentNullException("changes", Messages.ArgumentNull_Changes);
			}
			if (changes.Length == 0)
			{
				return;
			}

			string[] paths = Change.ToAbsolutePaths(changes); // TODO: check changes belong to workingtree

			using (PythonList filenames = PythonList.Create(paths))
			using (PythonNone old_tree = PythonNone.Create())
			using (PythonBool backups = PythonBool.Create(false))
			using (this.impl.InvokeMethod<PythonList>("revert", filenames, old_tree, backups)) // TODO: return conflicts
			{
			}
		}

		public void Update()
		{
			using (this.impl.InvokeMethod<PythonInteger>("update")) // TODO: return result
			{
			}
		}

		#endregion
	}
}