using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;

namespace Python
{
	public class PythonObject : SafeHandle, IEquatable<PythonObject>, IComparable<PythonObject>, IComparable, IDisposable //IEnumerable<PythonObject>
	{
		private static readonly Dictionary<IntPtr, PythonType> types = new Dictionary<IntPtr, PythonType>();

		public PythonObject()
			: base(IntPtr.Zero, true)
		{
		}

		public sealed override bool IsInvalid
		{
			get
			{
				return IsClosed || (base.handle == IntPtr.Zero);
			}
		}

		protected internal unsafe IntPtr NewReference()
		{
			NativeMethods.PyObject* op = (NativeMethods.PyObject*)base.handle;
			op->ob_refcnt = (IntPtr)((byte*)op->ob_refcnt + 1);
			return base.handle;
		}

		protected sealed override unsafe bool ReleaseHandle()
		{
			NativeMethods.PyObject* op = (NativeMethods.PyObject*)base.handle;
			IntPtr refcnt = (IntPtr)((byte*)op->ob_refcnt - 1);
			if (refcnt != IntPtr.Zero)
				op->ob_refcnt = refcnt;
			else
				NativeMethods.Py_DecRef(op);
			return true;
		}

		internal unsafe void SetHandle(NativeMethods.PyObject* op)
		{
			base.SetHandle((IntPtr)op);
		}

		public PythonType Type
		{
			get
			{
				PythonType type;

				lock (types)
				{
					if (!types.TryGetValue(base.handle, out type))
					{
						type = NativeMethods.PyObject_Type(this);
						types.Add(base.handle, type);
					}
				}

				return type;
			}
		}

		internal static PythonType GetType(ref IntPtr handle, string name)
		{
			PythonType type;

			lock (types)
			{
				if (IntPtr.Zero == handle || !types.TryGetValue(handle, out type))
				{
					using (PythonModule module = PythonException.ThrowIfInvalid(NativeMethods.PyImport_ImportModule("types")))
					{
						type = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsType(module, name));
						types.Add(type.handle, type);
						handle = type.handle;
					}
				}
			}

			return type;
		}

		/*
		public string Name
		{
			get
			{
				using (PythonString name = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrString<PythonString>(this, "__name__")))
				{
					return name.ToString();
				}
			}
		}

		public string Module
		{
			get
			{
				using (PythonString name = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrString<PythonString>(this, "__module__")))
				{
					return name.ToString();
				}
			}
		}
		*/







		public bool HasAttribute(string name)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}

			return NativeMethods.PyObject_HasAttrString(this, name) > 0;
		}

		public bool HasAttribute(PythonObject name)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}

			return NativeMethods.PyObject_HasAttr(this, name) > 0;
		}

		public PythonObject GetAttribute(string name)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrString(this, name));
		}

		public T GetAttribute<T>(string name) where T : PythonObject, new()
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrString<T>(this, name));
		}

		public PythonObject GetAttribute(string name, PythonObject @default)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.Clear(NativeMethods.PyObject_GetAttrString(this, name), @default);
		}

		public T GetAttribute<T>(string name, T @default) where T : PythonObject, new()
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.Clear(NativeMethods.PyObject_GetAttrString<T>(this, name), @default);
		}

		public PythonObject GetAttribute(PythonObject name)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttr(this, name));
		}

		public T GetAttribute<T>(PythonObject name) where T : PythonObject, new()
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttr<T>(this, name));
		}

		public PythonObject GetAttribute(PythonObject name, PythonObject @default)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.Clear(NativeMethods.PyObject_GetAttr(this, name), @default);
		}

		public T GetAttribute<T>(PythonObject name, T @default) where T : PythonObject, new()
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.Clear(NativeMethods.PyObject_GetAttr<T>(this, name), @default);
		}

		public void SetAttribute(string name, PythonObject value)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}
			if (null == value)
			{
				throw new ArgumentNullException("value");
			}

			PythonException.ThrowIfNegative(NativeMethods.PyObject_SetAttrString(this, name, value));
		}

		public void SetAttribute(PythonObject name, PythonObject value)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}
			if (null == value)
			{
				throw new ArgumentNullException("value");
			}

			PythonException.ThrowIfNegative(NativeMethods.PyObject_SetAttr(this, name, value));
		}

		public void DeleteAttribute(string name)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}

			PythonException.ThrowIfNegative(NativeMethods.PyObject_DelAttrString(this, name));
		}

		public void DeleteAttribute(PythonObject name)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}

			PythonException.ThrowIfNegative(NativeMethods.PyObject_DelAttr(this, name));
		}

		public PythonObject GetItem(PythonObject key)
		{
			if (null == key)
			{
				throw new ArgumentNullException("key");
			}

			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetItem(this, key));
		}

		public T GetItem<T>(PythonObject key) where T : PythonObject, new()
		{
			if (null == key)
			{
				throw new ArgumentNullException("key");
			}

			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetItem<T>(this, key));
		}

		public PythonObject GetItem(string key)
		{
			if (null == key)
			{
				throw new ArgumentNullException("key");
			}

			using (PythonString k = PythonException.ThrowIfInvalid(NativeMethods.PyString_FromStringAndSize(key, key.Length)))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetItem(this, k));
			}
		}

		public T GetItem<T>(string key) where T : PythonObject, new()
		{
			if (null == key)
			{
				throw new ArgumentNullException("key");
			}

			using (PythonString k = PythonException.ThrowIfInvalid(NativeMethods.PyString_FromStringAndSize(key, key.Length)))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetItem<T>(this, k));
			}
		}

		public PythonObject GetItem(int index)
		{
			using (PythonInteger k = PythonException.ThrowIfInvalid(NativeMethods.PyInt_FromInt32(index)))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetItem(this, k));
			}
		}

		public T GetItem<T>(int index) where T : PythonObject, new()
		{
			using (PythonInteger k = PythonException.ThrowIfInvalid(NativeMethods.PyInt_FromInt32(index)))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetItem<T>(this, k));
			}
		}

		public void SetItem(PythonObject key, PythonObject value)
		{
			if (null == key)
			{
				throw new ArgumentNullException("key");
			}
			if (null == value)
			{
				throw new ArgumentNullException("value");
			}

			PythonException.ThrowIfNegative(NativeMethods.PyObject_SetItem(this, key, value));
		}

		public void SetItem(string key, PythonObject value)
		{
			if (null == key)
			{
				throw new ArgumentNullException("key");
			}
			if (null == value)
			{
				throw new ArgumentNullException("value");
			}

			using (PythonString k = PythonException.ThrowIfInvalid(NativeMethods.PyString_FromStringAndSize(key, key.Length)))
			{
				PythonException.ThrowIfNegative(NativeMethods.PyObject_SetItem(this, k, value));
			}
		}

		public void SetItem(int index, PythonObject value)
		{
			if (null == value)
			{
				throw new ArgumentNullException("value");
			}

			using (PythonInteger k = PythonException.ThrowIfInvalid(NativeMethods.PyInt_FromInt32(index)))
			{
				PythonException.ThrowIfNegative(NativeMethods.PyObject_SetItem(this, k, value));
			}
		}

		public void DeleteItem(PythonObject key)
		{
			if (null == key)
			{
				throw new ArgumentNullException("key");
			}

			PythonException.ThrowIfNegative(NativeMethods.PyObject_DelItem(this, key));
		}

		public void DeleteItem(string key)
		{
			if (null == key)
			{
				throw new ArgumentNullException("key");
			}

			using (PythonString k = PythonException.ThrowIfInvalid(NativeMethods.PyString_FromStringAndSize(key, key.Length)))
			{
				PythonException.ThrowIfNegative(NativeMethods.PyObject_DelItem(this, k));
			}
		}

		public void DeleteItem(int index)
		{
			using (PythonInteger k = PythonException.ThrowIfInvalid(NativeMethods.PyInt_FromInt32(index)))
			{
				PythonException.ThrowIfNegative(NativeMethods.PyObject_DelItem(this, k));
			}
		}

		public PythonIterator<T> GetIterator<T>() where T : PythonObject, new()
		{
			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetIter<T>(this));
		}

		public PythonObject Invoke()
		{
			using (PythonTuple a = PythonTuple.Create())
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(this, a));
			}
		}

		public T Invoke<T>() where T : PythonObject, new()
		{
			using (PythonTuple a = PythonTuple.Create())
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(this, a));
			}
		}

		public PythonObject Invoke(PythonObject arg0)
		{
			if (null == arg0)
			{
				throw new ArgumentNullException("arg0");
			}

			using (PythonTuple a = PythonTuple.Create(arg0))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(this, a));
			}
		}

		public T Invoke<T>(PythonObject arg0) where T : PythonObject, new()
		{
			if (null == arg0)
			{
				throw new ArgumentNullException("arg0");
			}

			using (PythonTuple a = PythonTuple.Create(arg0))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(this, a));
			}
		}

		public PythonObject Invoke(PythonObject arg0, PythonObject arg1)
		{
			if (null == arg0)
			{
				throw new ArgumentNullException("arg0");
			}
			if (null == arg1)
			{
				throw new ArgumentNullException("arg1");
			}

			using (PythonTuple a = PythonTuple.Create(arg0, arg1))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(this, a));
			}
		}

		public T Invoke<T>(PythonObject arg0, PythonObject arg1) where T : PythonObject, new()
		{
			if (null == arg0)
			{
				throw new ArgumentNullException("arg0");
			}
			if (null == arg1)
			{
				throw new ArgumentNullException("arg1");
			}

			using (PythonTuple a = PythonTuple.Create(arg0, arg1))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(this, a));
			}
		}

		public PythonObject Invoke(PythonObject arg0, PythonObject arg1, PythonObject arg2)
		{
			if (null == arg0)
			{
				throw new ArgumentNullException("arg0");
			}
			if (null == arg1)
			{
				throw new ArgumentNullException("arg1");
			}
			if (null == arg2)
			{
				throw new ArgumentNullException("arg2");
			}

			using (PythonTuple a = PythonTuple.Create(arg0, arg1, arg2))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(this, a));
			}
		}

		public T Invoke<T>(PythonObject arg0, PythonObject arg1, PythonObject arg2) where T : PythonObject, new()
		{
			if (null == arg0)
			{
				throw new ArgumentNullException("arg0");
			}
			if (null == arg1)
			{
				throw new ArgumentNullException("arg1");
			}
			if (null == arg2)
			{
				throw new ArgumentNullException("arg2");
			}

			using (PythonTuple a = PythonTuple.Create(arg0, arg1, arg2))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(this, a));
			}
		}

		public PythonObject Invoke(params PythonObject[] args)
		{
			if (null == args)
			{
				throw new ArgumentNullException("args");
			}

			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(this, a));
			}
		}

		public T Invoke<T>(params PythonObject[] args) where T : PythonObject, new()
		{
			if (null == args)
			{
				throw new ArgumentNullException("args");
			}

			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(this, a));
			}
		}

		public PythonObject Invoke(PythonObject[] args, PythonDictionary kw)
		{
			if (null == args)
			{
				throw new ArgumentNullException("args");
			}
			if (null == kw)
			{
				throw new ArgumentNullException("kw");
			}

			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_Call(this, a, kw));
			}
		}

		public T Invoke<T>(PythonObject[] args, PythonDictionary kw) where T : PythonObject, new()
		{
			if (null == args)
			{
				throw new ArgumentNullException("args");
			}
			if (null == kw)
			{
				throw new ArgumentNullException("kw");
			}

			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_Call<T>(this, a, kw));
			}
		}

		public PythonObject InvokeMethod(string name)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create())
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(m, a));
			}
		}

		public T InvokeMethod<T>(string name) where T : PythonObject, new()
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create())
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(m, a));
			}
		}

		public PythonObject InvokeMethod(string name, PythonObject arg0)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}
			if (null == arg0)
			{
				throw new ArgumentNullException("arg0");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(arg0))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(m, a));
			}
		}

		public T InvokeMethod<T>(string name, PythonObject arg0) where T : PythonObject, new()
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}
			if (null == arg0)
			{
				throw new ArgumentNullException("arg0");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(arg0))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(m, a));
			}
		}

		public PythonObject InvokeMethod(string name, PythonObject arg0, PythonObject arg1)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}
			if (null == arg0)
			{
				throw new ArgumentNullException("arg0");
			}
			if (null == arg1)
			{
				throw new ArgumentNullException("arg1");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(arg0, arg1))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(m, a));
			}
		}

		public T InvokeMethod<T>(string name, PythonObject arg0, PythonObject arg1) where T : PythonObject, new()
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}
			if (null == arg0)
			{
				throw new ArgumentNullException("arg0");
			}
			if (null == arg1)
			{
				throw new ArgumentNullException("arg1");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(arg0, arg1))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(m, a));
			}
		}

		public PythonObject InvokeMethod(string name, PythonObject arg0, PythonObject arg1, PythonObject arg2)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}
			if (null == arg0)
			{
				throw new ArgumentNullException("arg0");
			}
			if (null == arg1)
			{
				throw new ArgumentNullException("arg1");
			}
			if (null == arg2)
			{
				throw new ArgumentNullException("arg2");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(arg0, arg1, arg2))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(m, a));
			}
		}

		public T InvokeMethod<T>(string name, PythonObject arg0, PythonObject arg1, PythonObject arg2) where T : PythonObject, new()
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}
			if (null == arg0)
			{
				throw new ArgumentNullException("arg0");
			}
			if (null == arg1)
			{
				throw new ArgumentNullException("arg1");
			}
			if (null == arg2)
			{
				throw new ArgumentNullException("arg2");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(arg0, arg1, arg2))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(m, a));
			}
		}

		public PythonObject InvokeMethod(string name, params PythonObject[] args)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}
			if (null == args)
			{
				throw new ArgumentNullException("args");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(m, a));
			}
		}

		public T InvokeMethod<T>(string name, params PythonObject[] args) where T : PythonObject, new()
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}
			if (null == args)
			{
				throw new ArgumentNullException("args");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(m, a));
			}
		}

		public PythonObject InvokeMethod(string name, PythonObject[] args, PythonDictionary kw)
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}
			if (null == args)
			{
				throw new ArgumentNullException("args");
			}
			if (null == kw)
			{
				throw new ArgumentNullException("kw");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_Call(m, a, kw));
			}
		}

		public T InvokeMethod<T>(string name, PythonObject[] args, PythonDictionary kw) where T : PythonObject, new()
		{
			if (null == name)
			{
				throw new ArgumentNullException("name");
			}
			if (null == args)
			{
				throw new ArgumentNullException("args");
			}
			if (null == kw)
			{
				throw new ArgumentNullException("kw");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_Call<T>(m, a, kw));
			}
		}

		public bool IsInstanceOf(PythonType type)
		{
			if (null == type)
			{
				throw new ArgumentNullException("type");
			}

			return type.IsSubtypeOf(this.Type);
		}

		public bool IsCallable
		{
			get
			{
				return NativeMethods.PyCallable_Check(this) > 0;
			}
		}

		public bool IsTrue
		{
			get
			{
				return PythonException.ThrowIfNegative(NativeMethods.PyObject_IsTrue(this)) > 0;
			}
		}

		public bool IsFalse
		{
			get
			{
				return PythonException.ThrowIfNegative(NativeMethods.PyObject_Not(this)) > 0;
			}
		}

		public PythonList Dir()
		{
			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_Dir(this));
		}

		public string Repr()
		{
			using (PythonString repr = PythonException.ThrowIfInvalid(NativeMethods.PyObject_Repr(this)))
			{
				return repr.ToString();
			}
		}

		public override string ToString()
		{
			using (PythonString unicode = PythonException.ThrowIfInvalid(NativeMethods.PyObject_Unicode(this)))
			{
				return unicode.ToString();
			}
		}

		#region IEquatable Members

		public override int GetHashCode()
		{
			using (PythonInteger o = PythonException.ThrowIfInvalid(NativeMethods.PyObject_Hash(this)))
			{
				return o.ToInt32();
			}
		}

		public override bool Equals(object obj)
		{
			if (null == obj)
			{
				return false;
			}

			if ((object)this == obj)
			{
				return true;
			}

			return CompareTo(obj as PythonObject) == 0;
		}

		public bool Equals(PythonObject other)
		{
			return CompareTo(other) == 0;
		}

		#endregion

		#region IComparable Members

		public int CompareTo(object obj)
		{
			if (null == obj)
			{
				return 1;
			}

			if ((object)this == obj)
			{
				return 0;
			}

			PythonObject other = obj as PythonObject;

			if (null == other)
			{
				throw new ArgumentException();
			}

			return CompareTo(other);
		}

		public int CompareTo(PythonObject other)
		{
			if (null == other)
			{
				return 1;
			}

			int value = NativeMethods.PyObject_Compare(this, other);
			PythonException.ThrowIfErrorOccurred();
			return value;
		}

		#endregion

		//#region IEnumerable Members

		//public IEnumerator<PythonObject> GetEnumerator()
		//{
		//    return GetIterator<PythonObject>();
		//}

		//System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		//{
		//    return GetIterator<PythonObject>();
		//}

		//#endregion
	}
}